# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)

import numpy as np

from .entity_with_sources import EntityWithSources
from ..tag import TagMixin
from ..value import DataType
from .data_array import DataArray
from .data_view import DataView
from .feature import Feature
from .exceptions import (OutOfBounds, IncompatibleDimensions,
                         UninitializedEntity, InvalidUnit)
from ..dimension_type import DimensionType
from ..link_type import LinkType
from . import util


class BaseTag(EntityWithSources):
    """
    Base class for Tag and MultiTag
    """

    @property
    def units(self):
        """
        Property containing the units of the tag. The tag must provide a
        unit for each dimension of the position or extent vector.
        This is a read-write property.

        :type: list of str
        """
        return tuple(u.decode() if isinstance(u, bytes) else u
                     for u in self._h5group.get_data("units"))

    @units.setter
    def units(self, units):
        if not units:
            if self._h5group.has_data("units"):
                del self._h5group["units"]
        else:
            sanitized = []
            for u in units:
                util.check_attr_type(u, str)
                u = util.units.sanitizer(u)
                sanitized.append(u)

            dtype = DataType.String
            self._h5group.write_data("units", sanitized, dtype)
            self.force_updated_at()

    def _add_reference_by_id(self, id_or_name):
        if id_or_name not in self._parent.data_arrays:
            cls = type(self).__name__
            raise RuntimeError("{}._add_reference_by_id: "
                               "Reference not found in Block!".format(cls))
        target = self._parent.data_arrays[id_or_name]
        references = self._h5group.open_group("references")
        references.create_link(target, target.id)

    def _has_reference_by_id(self, id_or_name):
        references = self._h5group.open_group("references")
        return references.has_by_id(id_or_name)

    def _reference_count(self):
        return len(self._h5group.open_group("references"))

    def _get_reference_by_id(self, id_or_name):
        references = self._h5group.open_group("references")
        if util.is_uuid(id_or_name):
            id_ = id_or_name
        else:
            id_ = self._parent.data_arrays[id_or_name].id
        return DataArray(self._parent, references.get_by_id(id_))

    def _get_reference_by_pos(self, pos):
        references = self._h5group.open_group("references")
        return DataArray(self, references.get_by_pos(pos))

    def _delete_reference_by_id(self, id_):
        references = self._h5group.open_group("references")
        references.delete(id_)

    def create_feature(self, data, link_type):
        """
        Create a new feature.

        :param data: The data array of this feature.
        :type data: DataArray
        :param link_type: The link type of this feature.
        :type link_type: LinkType

        :returns: The created feature object.
        :rtype: Feature
        """
        features = self._h5group.open_group("features")
        feat = Feature._create_new(self, features, data, link_type)
        return feat

    def _has_feature_by_id(self, id_or_name):
        features = self._h5group.open_group("features")
        return features.has_by_id(id_or_name)

    def _feature_count(self):
        return len(self._h5group.open_group("features"))

    def _get_feature_by_id(self, id_or_name):
        features = self._h5group.open_group("features")
        try:
            return Feature(self, features.get_by_id(id_or_name))
        except ValueError:
            for feat in self.features:
                if feat.data.id == id_or_name or feat.data.name == id_or_name:
                    return feat

    def _get_feature_by_pos(self, pos):
        features = self._h5group.open_group("features")
        return Feature(self, features.get_by_pos(pos))

    def _delete_feature_by_id(self, id_):
        features = self._h5group.open_group("features")
        features.delete(id_)

    @classmethod
    def _position_and_extent_in_data(cls, data, offset, count):
        pos = tuple(np.add(offset, count) - 1)
        return cls._position_in_data(data, pos)

    @staticmethod
    def _position_in_data(data, pos):
        dasize = data.data_extent
        return np.all(np.less(pos, dasize))

    @staticmethod
    def _pos_to_idx(pos, unit, dim):
        dimtype = dim.dimension_type
        if dimtype == DimensionType.Set:
            dimunit = None
        else:
            dimunit = dim.unit
        scaling = 1.0
        if dimtype == DimensionType.Sample:
            if not dimunit and unit is not None:
                raise IncompatibleDimensions(
                    "Units of position and SampledDimension "
                    "must both be given!",
                    "Tag._pos_to_idx"
                )
            if dimunit and unit is not None:
                try:
                    scaling = util.units.scaling(unit, dimunit)
                except InvalidUnit:
                    raise IncompatibleDimensions(
                        "Cannot apply a position with unit to a SetDimension",
                        "Tag._pos_to_idx"
                    )

            index = dim.index_of(pos * scaling)
        elif dimtype == DimensionType.Set:
            if unit and unit != "none":
                raise IncompatibleDimensions(
                    "Cannot apply a position with unit to a SetDimension",
                    "Tag._pos_to_idx"
                )
            index = round(pos)
            nlabels = len(dim.labels)
            if nlabels and index > nlabels:
                raise OutOfBounds("Position is out of bounds in SetDimension",
                                  pos)
        else:  # dimtype == DimensionType.Range:
            if dimunit and unit is not None:
                try:
                    scaling = util.units.scaling(unit, dimunit)
                except InvalidUnit:
                    raise IncompatibleDimensions(
                        "Provided units are not scalable!",
                        "Tag._pos_to_idx"
                    )
            index = dim.index_of(pos * scaling)

        return int(index)


class Tag(BaseTag, TagMixin):

    def __init__(self, nixparent, h5group):
        super(Tag, self).__init__(nixparent, h5group)

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_, position):
        newentity = super(Tag, cls)._create_new(nixparent, h5parent,
                                                name, type_)
        newentity.position = position
        return newentity

    @property
    def position(self):
        """
        The position defined by the tag. This is a read-write property.

        :type: list of float
        """
        return tuple(self._h5group.get_data("position"))

    @position.setter
    def position(self, pos):
        if pos is None or len(pos) == 0:
            if self._h5group.has_data("position"):
                del self._h5group["position"]
        else:
            dtype = DataType.Double
            self._h5group.write_data("position", pos, dtype)

    @property
    def extent(self):
        """
        The extent defined by the tag. This is an optional read-write
        property and may be set to None.

        :type: list of float
        """
        return tuple(self._h5group.get_data("extent"))

    @extent.setter
    def extent(self, ext):
        if ext is None or len(ext) == 0:
            if self._h5group.has_data("extent"):
                del self._h5group["extent"]
        else:
            dtype = DataType.Double
            self._h5group.write_data("extent", ext, dtype)

    def _get_offset_and_count(self, data):
        offset = []
        count = []
        position = self.position
        extent = self.extent
        for idx in range(len(position)):
            dim = data.dimensions[idx]
            pos = position[idx]
            if self.units:
                unit = self.units[idx]
            else:
                unit = None
            o = self._pos_to_idx(position[idx], unit, dim)
            offset.append(o)
            if idx < len(extent):
                ext = extent[idx]
                c = self._pos_to_idx(pos + ext, unit, dim) - o
                count.append(c if c > 1 else 1)
            else:
                count.append(1)
        return tuple(offset), tuple(count)

    def retrieve_data(self, refidx):
        references = self.references
        position = self.position
        extent = self.extent
        if len(references) == 0:
            raise OutOfBounds("There are no references in this tag!")

        if isinstance(refidx, int) and refidx >= len(references):
            raise OutOfBounds("Reference index out of bounds.")

        ref = references[refidx]
        dimcount = len(ref.dimensions)
        if (len(position) != dimcount) or (len(extent) > 0 and
                                           len(extent) != dimcount):
            raise IncompatibleDimensions(
                "Number of dimensions in position or extent do not match "
                "dimensionality of data",
                "Tag.retrieve_data")

        offset, count = self._get_offset_and_count(ref)

        if not self._position_and_extent_in_data(ref, offset, count):
            raise OutOfBounds("References data slice out of the extent of the "
                              "DataArray!")
        return DataView(ref, count, offset)

    def retrieve_feature_data(self, featidx):
        if self._feature_count() == 0:
            raise OutOfBounds(
                "There are no features associated with this tag!"
            )

        try:
            feat = self.features[featidx]
        except KeyError:
            feat = None
            for f in self.features:
                if f.data.name == featidx or f.data.id == featidx:
                    feat = f
                    break
            if feat is None:
                raise
        da = feat.data
        if da is None:
            raise UninitializedEntity()
        if feat.link_type == LinkType.Tagged:
            offset, count = self._get_offset_and_count(da)
            if not self._position_and_extent_in_data(da, offset, count):
                raise OutOfBounds("Requested data slice out of the extent "
                                  "of the Feature!")
            return DataView(da, count, offset)
        # For untagged and indexed return the full data
        count = da.data_extent
        offset = (0,) * len(count)
        return DataView(da, count, offset)



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)

from collections import Sequence

from ..section import SectionMixin
from .entity import NamedEntity
from .property import Property
from . import util
from . import exceptions


class Section(NamedEntity, SectionMixin):

    def __init__(self, nixparent, h5group):
        super(Section, self).__init__(nixparent, h5group)
        self._sec_parent = None

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_):
        newentity = super(Section, cls)._create_new(nixparent, h5parent,
                                                    name, type_)
        return newentity

    # Section
    def create_section(self, name, type_):
        """
        Creates a new subsection that is a child of this section entity.

        :param name: The name of the section to create.
        :type name: str
        :param type_: The type of the section.
        :type type_: str

        :returns: The newly created section.
        :rtype: Section
        """
        util.check_entity_name_and_type(name, type_)
        sections = self._h5group.open_group("sections", True)
        if name in sections:
            raise exceptions.DuplicateName("create_section")
        sec = Section._create_new(self, sections, name, type_)
        sec._sec_parent = self
        return sec

    def _get_section_by_id(self, id_or_name):
        sections = self._h5group.open_group("sections")
        return Section(self, sections.get_by_id_or_name(id_or_name))

    def _get_section_by_pos(self, pos):
        sections = self._h5group.open_group("sections")
        return Section(self, sections.get_by_pos(pos))

    def _delete_section_by_id(self, id_):
        sections = self._h5group.open_group("sections")
        sections.delete(id_)

    def _section_count(self):
        return len(self._h5group.open_group("sections"))

    # Property
    def create_property(self, name, values):
        """
        Add a new property to the section.

        :param name: The name of the property to create.
        :type name: str
        :param values: The values of the property.
        :type values: list of Value

        :returns: The newly created property.
        :rtype: Property
        """
        properties = self._h5group.open_group("properties", True)
        if name in properties:
            raise exceptions.DuplicateName("create_property")
        if isinstance(values, type):
            dtype = values
            values = []
        else:
            if isinstance(values, Sequence):
                dtype = values[0].data_type
            else:
                dtype = values.data_type
                values = [values]
        prop = Property._create_new(self, properties, name, dtype)
        prop.values = values
        return prop

    def has_property_by_name(self, name):
        """
        Checks whether a section has a property with a certain name.

        :param name: The name to check.
        :type name: str

        :returns: True if the section has a property with the given name,
                  False otherwise.
        :rtype: bool
        """
        properties = self._h5group.open_group("properties")
        return properties.has_by_id(name)

    def get_property_by_name(self, name):
        """
        Get a property by its name.

        :param name: The name to check.
        :type name: str

        :returns: The property with the given name.
        :rtype: Property
        """
        properties = self._h5group.open_group("properties")
        try:
            p = Property(self, properties.get_by_name(name))
        except ValueError:
            p = None
        return p

    def _get_property_by_id(self, id_or_name):
        properties = self._h5group.open_group("properties")
        return Property(self, properties.get_by_id_or_name(id_or_name))

    def _get_property_by_pos(self, pos):
        properties = self._h5group.open_group("properties")
        return Property(self, properties.get_by_pos(pos))

    def _delete_property_by_id(self, id_):
        properties = self._h5group.open_group("properties")
        properties.delete(id_)

    def _property_count(self):
        return len(self._h5group.open_group("properties"))

    @property
    def link(self):
        """
        Link to another section. If a section is linked to another section,
        the linking section inherits all properties from the target section.
        This is an optional read-write property and may be set to None.

        :type: Section
        """
        if "link" not in self._h5group:
            return None
        else:
            return Section(self, self._h5group.open_group("link"))

    @link.setter
    def link(self, id_or_sec):
        if id_or_sec is None:
            self._h5group.delete("link")
        if isinstance(id_or_sec, Section):
            sec = id_or_sec
        else:
            rootsec = Section(self, self._h5group.h5root)
            sec = rootsec.find_sections(filtr=lambda x: x.id == id_or_sec)

        self._h5group.create_link(sec, "link")

    def inherited_properties(self):
        properties = self._h5group.open_group("properties")
        inhprops = [Property(self, h5prop) for h5prop in properties]
        if self.link:
            inhprops.append(self.link.inherited_properties())
        return inhprops

    @property
    def mapping(self):
        return self._h5group.get_attr("mapping")

    @mapping.setter
    def mapping(self, m):
        """
        The mapping information of the section.
        This is an optional read-write property and may be set to None.

        :type: str
        """
        util.check_attr_type(m, str)
        self._h5group.set_attr("mapping", m)

    @property
    def repository(self):
        """
        URL to the terminology repository the section is associated with.
        This is an optional read-write property and may be set to None.

        :type: str
        """
        return self._h5group.get_attr("repository")

    @repository.setter
    def repository(self, r):
        util.check_attr_type(r, str)
        self._h5group.set_attr("repository", r)

    @property
    def parent(self):
        """
        The parent section. This is a read-only property. For root sections
        this property is always None.

        Accessing this property can be slow when the metadata tree is large.

        :type: Section
        """
        if self._sec_parent is not None:
            return self._sec_parent
        rootmd = self._h5group.file.open_group("metadata")
        # Assuming most metadata trees are shallow---doing BFS
        sections = [Section(None, sg) for sg in rootmd]
        if self in sections:
            # Top-level section
            return None

        while sections:
            sect = sections.pop(0)
            if self in sect.sections:
                self._sec_parent = sect
                return sect
            sections.extend(sect.sections)

        return None

    @property
    def file(self):
        """
        Root file object.

        :type: File
        """
        par = self._parent
        while isinstance(par, NamedEntity):
            par = par._parent
        return par

    @property
    def referring_objects(self):
        objs = []
        objs.extend(self.referring_blocks)
        objs.extend(self.referring_groups)
        objs.extend(self.referring_data_arrays)
        objs.extend(self.referring_tags)
        objs.extend(self.referring_multi_tags)
        objs.extend(self.referring_sources)
        return objs

    @property
    def referring_blocks(self):
        f = self.file
        return list(blk for blk in f.blocks if blk.metadata.id == self.id)

    @property
    def referring_groups(self):
        f = self.file
        groups = []
        for blk in f.blocks:
            groups.extend(grp for grp in blk.groups
                          if grp.metadata.id == self.id)
        return groups

    @property
    def referring_data_arrays(self):
        f = self.file
        data_arrays = []
        for blk in f.blocks:
            data_arrays.extend(da for da in blk.data_arrays
                               if da.metadata.id == self.id)
        return data_arrays

    @property
    def referring_tags(self):
        f = self.file
        tags = []
        for blk in f.blocks:
            tags.extend(tg for tg in blk.tags
                        if tg.metadata.id == self.id)
        return tags

    @property
    def referring_multi_tags(self):
        f = self.file
        multi_tags = []
        for blk in f.blocks:
            multi_tags.extend(mt for mt in blk.multi_tags
                              if mt.metadata.id == self.id)
        return multi_tags

    @property
    def referring_sources(self):
        f = self.file
        sources = []
        for blk in f.blocks:
            sources.extend(src for src in blk.sources
                           if src.metadata.id == self.id)
        return sources

    def pprint(self, current_depth=0, max_depth=1, indent=2, max_length=80):
        spaces = " " * (current_depth * indent)
        sec_str = "%s  %s [%s] --- %s" % (spaces, self.name, self.type, self.id)
        print(sec_str)
        for p in self.props:
            p.pprint(current_depth, indent, max_length=max_length)
        if max_depth == -1 or current_depth < max_depth:
            for s in self.sections:
                s.pprint(current_depth+1, max_depth=max_depth, indent=indent, max_length=max_length)
        elif max_depth == current_depth:
            child_sec_indent = spaces+" "*indent
            more_indent =  spaces+" "*(current_depth+2*indent)
            for s in self.sections:
                print("%s  %s [%s] --- %s\n%s[...]" % (child_sec_indent, s.name, s.type,
                                                        s.id, more_indent))




from .property import Property
from .section import Section
from .source import Source
from .data_array import DataArray
from .dimensions import RangeDimension, SetDimension, SampledDimension
from .feature import Feature
from .tag import Tag
from .multi_tag import MultiTag
from .group import Group
from .block import Block
from .file import File, FileMode



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import sys
from numbers import Integral

from nixio.dimension_type import DimensionType

import numpy as np


class DataArrayMixin(object):

    @property
    def data(self):
        """
        DEPRECATED DO NOT USE ANYMORE! Returns self

        :type: :class:`~nixio.data_array.DataArray`
        """
        import warnings
        warnings.warn("Call to deprecated property DataArray.data",
                      category=DeprecationWarning)
        return self

    @property
    def dimensions(self):
        """
        A property containing all dimensions of a DataArray. Dimensions can be
        obtained via their index. Adding dimensions is done using the
        respective append methods for dimension descriptors.
        This is a read only attribute.

        :type: ProxyList of dimension descriptors.
        """
        if not hasattr(self, "_dimensions"):
            setattr(self, "_dimensions", DimensionProxyList(self))
        return self._dimensions

    def __eq__(self, other):
        if hasattr(other, "id"):
            return self.id == other.id
        else:
            return False

    def __hash__(self):
        """
        Overwriting method __eq__ blocks inheritance of __hash__ in Python 3
        hash has to be either explicitly inherited from parent class,
        implemented or escaped
        """
        return hash(self.id)


class SetDimensionMixin(object):

    dimension_type = DimensionType.Set


class RangeDimensionMixin(object):

    dimension_type = DimensionType.Range


class SampleDimensionMixin(object):

    dimension_type = DimensionType.Sample


class DimensionProxyList(object):
    """
    List proxy for the dimensions of a data array.
    """

    def __init__(self, obj):
        self.__obj = obj

    def __len__(self):
        return self.__obj._dimension_count()

    def __getitem__(self, key):
        if isinstance(key, int):
            length = self.__obj._dimension_count()

            if key < 0:
                key = length + key

            if key >= length or key < 0:
                raise KeyError("Index out of bounds: " + str(key))

            return self.__obj._get_dimension_by_pos(key + 1)
        else:
            raise TypeError("The key must be an int but was: " + type(key))

    def __iter__(self):
        for i in range(0, len(self)):
            yield self.__obj._get_dimension_by_pos(i + 1)

    def __str__(self):
        str_list = [str(e) for e in list(self)]
        return "[" + ", ".join(str_list) + "]"

    def __repr__(self):
        return str(self)


class DataSetMixin(object):
    """
    Data IO object for DataArray.
    """

    def __array__(self):
        raw = np.empty(self.shape, dtype=self.dtype)
        self.read_direct(raw)
        return raw

    def __getitem__(self, index):
        index = self.__index_to_tuple(index)
        if len(index) < 1:
            return np.array(self)
        # if we got to here we have a tuple with len >= 1
        count, offset, shape = self.__tuple_to_count_offset_shape(index)
        raw = np.empty(shape, dtype=self.dtype)

        if hasattr(self, "polynom_coefficients") and self.polynom_coefficients:
            # if there are coefficients, convert the dtype of the returned data
            # array to double
            raw.dtype = np.float64
        self._read_data(raw, count, offset)

        return raw

    def __setitem__(self, index, value):
        index = self.__index_to_tuple(index)
        if len(index) < 1:
            shape = self.shape
            count, offset = shape, tuple([0]*len(shape))
        else:
            count, offset, _ = self.__tuple_to_count_offset_shape(index)

        # NB: np.ascontiguousarray does not copy the array if it is
        # already in c-contiguous form
        raw = np.ascontiguousarray(value)
        self._write_data(raw, count, offset)

    def __len__(self):
        s = self.len()

        # PyObject_Size returns a Py_ssize_t, which is the same as the
        # systems size_t type but signed, i.e. ssize_t. (cf. PEP 0353)
        # The maximum positive integer that Py_ssize_t can hold is
        # exposed via sys.maxsize.
        # Since self.shape can contain longs we need to check for that
        if s > sys.maxsize:
            estr = ("DataSet's shape[0] is too big for Python's __len__. "
                    "Use DataSet.len() instead")
            raise OverflowError(estr)
        return s

    def __iter__(self):
        for idx in range(self.len()):
            yield self[idx]

    def len(self):
        """
        Length of the first dimension. Equivalent to `DataSet.shape[0]`.

        :type: int or long
        """
        return self.shape[0]

    @property
    def shape(self):
        """
        :type: tuple of data array dimensions.
        """
        return self.data_extent

    @property
    def size(self):
        """
        Number of elements in the DataSet, i.e. the product of the
        elements in :attr:`~nixio.data_array.DataSet.shape`.

        :type: int
        """
        return np.prod(self.shape)

    @property
    def dtype(self):
        """
        :type: :class:`numpy.dtype` object holding type infromation about
               the data stored in the DataSet.
        """
        return np.dtype(self._get_dtype())

    def write_direct(self, data):
        """
        Directly write all of ``data`` to the
        :class:`~nixio.data_array.DataSet`.  The supplied data must be a
        :class:`numpy.ndarray` that matches the DataSet's shape and must have
        C-style contiguous memory layout (see :attr:`numpy.ndarray.flags` and
        :class:`~numpy.ndarray` for more information).

        :param data: The array which contents is being written
        :type data: :class:`numpy.ndarray`
        """
        self._write_data(data, (), ())

    def read_direct(self, data):
        """
        Directly read all data stored in the :class:`~nixio.data_array.DataSet`
        into ``data``. The supplied data must be a :class:`numpy.ndarray` that
        matches the DataSet's shape, must have C-style contiguous memory layout
        and must be writeable (see :attr:`numpy.ndarray.flags` and
        :class:`~numpy.ndarray` for more information).

        :param data: The array where data is being read into
        :type data: :class:`numpy.ndarray`
        """

        self._read_data(data, (), ())

    def append(self, data, axis=0):
        """
        Append ``data`` to the DataSet along the ``axis`` specified.

        :param data: The data to append. Shape must agree except for the
        specified axis
        :param axis: Along which axis to append the data to
        """
        data = np.ascontiguousarray(data)

        if len(self.shape) != len(data.shape):
            raise ValueError(
                "Data and DataArray must have the same dimensionality"
            )

        if any([s != ds for i, (s, ds) in
                enumerate(zip(self.shape, data.shape)) if i != axis]):
            raise ValueError("Shape of data and shape of DataArray must match "
                             "in all dimension but axis!")

        offset = tuple(0 if i != axis else x for i, x in enumerate(self.shape))
        count = data.shape
        enlarge = tuple(self.shape[i] + (0 if i != axis else x)
                        for i, x in enumerate(data.shape))
        self.data_extent = enlarge
        self._write_data(data, count, offset)

    @staticmethod
    def __index_to_tuple(index):
        if isinstance(index, tuple):
            return index
        elif isinstance(index, Integral) or isinstance(index, slice):
            return (index, )
        elif isinstance(index, type(Ellipsis)):
            return ()
        else:
            raise IndexError("Unsupported index")

    @staticmethod
    def __complete_slices(shape, index):
        if isinstance(index, slice):
            if index.step is not None:
                raise IndexError('Invalid index, stepping unsupported')
            start = index.start
            stop = index.stop
            if start is None:
                start = 0
            elif start < 0:
                start = shape + start
            if stop is None:
                stop = shape
            elif stop < 0:
                stop = shape + stop
            index = slice(start, stop, index.step)
        elif isinstance(index, Integral):
            if index < 0:
                index = shape + index
                index = slice(index, index+1)
            else:
                index = slice(index, index+1)
        elif index is None:
            index = slice(0, shape)
        else:
            raise IndexError('Invalid index')
        return index

    @staticmethod
    def __fill_none(shape, index, to_replace=1):
        size = len(shape) - len(index) + to_replace
        return tuple([None] * size)

    def __tuple_to_count_offset_shape(self, index):
        # precondition: type(index) == tuple and len(index) >= 1
        fill_none = self.__fill_none
        shape = self.shape

        if index[0] is Ellipsis:
            index = fill_none(shape, index) + index[1:]
        if index[-1] is Ellipsis:
            # if we have a trailing ellipsis we just cut it away
            # and let complete_slices do the right thing
            index = index[:-1]

        # here we handle Ellipsis in the middle of the tuple
        # we *can* only handle one, if there are more, then
        # __complete_slices will raise a InvalidIndex error
        pos = index.index(Ellipsis) if Ellipsis in index else -1
        if pos > -1:
            index = index[:pos] + fill_none(shape, index) + index[pos+1:]

        # in python3 map does not work with None therefore if
        # len(shape) != len(index) we wont get the expected
        # result. We therefore need to fill up the missing values
        index = index + fill_none(shape, index, to_replace=0)

        completed = list(map(self.__complete_slices, shape, index))
        combined = list(map(lambda s: (s.start, s.stop), completed))
        count = tuple(x[1] - x[0] for x in combined)
        offset = [x for x in zip(*combined)][0]

        # drop all indices from count that came from single ints
        # NB: special case when we only have ints, e.g. (int, ) then
        # we get back the empty tuple and this is what we want,
        # because it indicates a scalar result
        squeezed = map(lambda i, c: c if type(i) != int
                       else None, index, count)
        shape = list(filter(lambda x: x is not None, squeezed))

        return count, offset, shape



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Copyright © 2014 German Neuroinformatics Node (G-Node)

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted under the terms of the BSD License. See
 LICENSE file in the root of the Project.

 Author: Jan Grewe <jan.grewe@g-node.org>

 This tutorial shows how to store image data in nix-files.
 See https://github.com/G-node/nix/wiki for more information.

 We use the "Lenna" image in this tutorial.

 "Lenna" by Original full portrait: "Playmate of the Month". Playboy
 Magazine. November 1972, photographed by Dwight Hooker.This 512x512
 electronic/mechanical scan of a section of the full portrait:
 Alexander Sawchuk and two others[1] - The USC-SIPI image
 database. Via Wikipedia -
 http://en.wikipedia.org/wiki/File:Lenna.png#mediaviewer/File:Lenna.png

"""

import nixio as nix
import numpy as np
import Image as img


def load_image():
    image = img.open('lenna.png')
    pix = np.array(image)
    channels = list(image.mode)
    return pix, channels


def draw_rect(img_data, position, extent):
    img_data[position[0]:position[0] + extent[0], position[1], :] = 255
    img_data[position[0]:position[0] + extent[0], position[1]+extent[1], :] = 255
    img_data[position[0], position[1]:position[1] + extent[1], :] = 255
    img_data[position[0] + extent[0], position[1]:position[1] + extent[1], :] = 255


def plot_data(tag):
    data_array = tag.references[0]
    img_data = np.zeros(data_array.data.shape)
    data_array.data.read_direct(img_data)
    img_data = np.array(img_data, dtype='uint8')
    pos = tag.position
    ext = tag.extent
    draw_rect(img_data, pos, ext)
    new_img = img.fromarray(img_data)
    new_img.show()


if __name__ == '__main__':
    img_data, channels = load_image()
    # create a new file overwriting any existing content
    file_name = 'single_roi.h5'
    file = nix.File.open(file_name, nix.FileMode.Overwrite)

    # create a 'Block' that represents a grouping object. Here, the recording session.
    # it gets a name and a type
    block = file.create_block("block name", "nix.session")

    # create a 'DataArray' to take the sinewave, add some information about the signal
    data = block.create_data_array("lenna", "nix.image.rgb", data=img_data)
    # add descriptors for width, height and channels
    height_dim = data.append_sampled_dimension(1)
    height_dim.label = "height"
    width_dim = data.append_sampled_dimension(1)
    width_dim.label = "width"
    color_dim = data.append_set_dimension()
    color_dim.labels = channels

    # create a Tag, position and extent must be 3-D since the data is 3-D
    position = [250, 250, 0]
    extent = [30, 100, 3]
    tag = block.create_tag('Region of interest', 'nix.roi', position)
    tag.extent = extent
    tag.references.append(data)

    # let's plot the data from the stored information
    plot_data(tag)
    file.close()



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import nixio.core

from nixio.file import FileMixin
from nixio.block import BlockMixin
from nixio.section import SectionMixin
from nixio.property import PropertyMixin
from nixio.group import GroupMixin
from nixio.data_array import (DataSetMixin, DataArrayMixin,
                              SetDimensionMixin, RangeDimensionMixin,
                              SampleDimensionMixin)
from nixio.source import SourceMixin
from nixio.tag import TagMixin
from nixio.multi_tag import MultiTagMixin
from nixio.entity_with_sources import EntityWithSourcesMixin

excludes = ("__module__", "__metaclass__", "__dict__", "__doc__")


def inject(bases, dct):
    """
    Does monkey patching to the classes in 'bases' by adding
    methods from the given dict 'dct'.
    """
    for base in bases:
        for k, v in dct.items():
            if k not in excludes:
                setattr(base, k, v)


inject((nixio.core.File,),  dict(FileMixin.__dict__))
inject((nixio.core.Section,), dict(SectionMixin.__dict__))
inject((nixio.core.Property,), dict(PropertyMixin.__dict__))
inject((nixio.core.Block,), dict(BlockMixin.__dict__))
inject((nixio.core.Group,), dict(GroupMixin.__dict__))
inject((nixio.core.DataArray,), dict(DataArrayMixin.__dict__))
inject((nixio.core.DataSet,), dict(DataSetMixin.__dict__))
inject((nixio.core.Tag,), dict(TagMixin.__dict__))
inject((nixio.core.MultiTag,), dict(MultiTagMixin.__dict__))
inject((nixio.core.Source,), dict(SourceMixin.__dict__))
inject((nixio.core.SetDimension,), dict(SetDimensionMixin.__dict__))
inject((nixio.core.RangeDimension,), dict(RangeDimensionMixin.__dict__))
inject((nixio.core.SampledDimension,), dict(SampleDimensionMixin.__dict__))
inject((nixio.core.Group,), dict(EntityWithSourcesMixin.__dict__))
inject((nixio.core.DataArray,), dict(EntityWithSourcesMixin.__dict__))
inject((nixio.core.Tag,), dict(EntityWithSourcesMixin.__dict__))
inject((nixio.core.MultiTag,), dict(EntityWithSourcesMixin.__dict__))



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
# -*- coding: utf-8 -*-
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)

import re
from collections import Sequence
from ..exceptions import InvalidUnit


strings = (str, bytes)
try:
    strings += (basestring,)
except NameError:
    pass

# Base32hex alphabet (RFC 4648)
ID_ALPHABET = "0123456789abcdefghijklmnopqrstuv"
# Unit scaling, SI only, substitutions for micro and ohm...
PREFIXES = "(Y|Z|E|P|T|G|M|k|h|da|d|c|m|u|n|p|f|a|z|y)"
UNITS = ("(m|g|s|A|K|mol|cd|Hz|N|Pa|J|W|C|V|F|S|Wb|T|H|lm|lx|Bq|Gy|Sv|kat|l|L|"
         "Ohm|%|dB|rad)")
POWER = "(\\^[+-]?[1-9]\\d*)"
PREFIX_FACTORS = {"y": 1.0e-24,
                  "z": 1.0e-21,
                  "a": 1.0e-18,
                  "f": 1.0e-15,
                  "p": 1.0e-12,
                  "n": 1.0e-9,
                  "u": 1.0e-6,
                  "m": 1.0e-3,
                  "c": 1.0e-2,
                  "d": 1.0e-1,
                  "da": 1.0e1,
                  "h": 1.0e2,
                  "k": 1.0e3,
                  "M": 1.0e6,
                  "G": 1.0e9,
                  "T": 1.0e12,
                  "P": 1.0e15,
                  "E": 1.0e18,
                  "Z": 1.0e21,
                  "Y": 1.0e24}


def sanitizer(unit):
    """
    Sanitizes a unit string. That is, it is de-blanked, and mu and µ symbols
    are changed to u for micro.

    :param unit: The unit that needs to be sanitized.

    :returns: the sanitized unit.
    :rtype: str
    """
    # micro = "\u03bc"
    micro = "µ"
    # mugr = "\u00b5"
    mugr = "μ"
    return unit.replace(" ", "").replace("mu", "u").\
        replace(micro, "u").replace(mugr, "u")


def is_si(unit):
    """
    Determines whether a unit is a recognized SI unit.

    :param unit: The unit that needs to be checked.

    :returns: True if the unit is an SI unit, false otherwise.
    :rtype: bool
    """
    return unit and (is_atomic(unit) or is_compound(unit))


def is_atomic(unit):
    """
    Checked whether a unit string represents an atomic si unit, i.e. not a
    combination.

    :param unit: The unit to be checked.

    :returns: True if unit is atomic, False otherwise.
    :rtype: bool
    """
    atomic_unit = re.compile(
        "^{prefix}?{unit}{power}?$".format(prefix=PREFIXES,
                                           unit=UNITS,
                                           power=POWER)
    )
    return atomic_unit.match(unit)


def is_compound(unit):
    """
    Checks whether a unit string represents a combination of SI units.

    :param unit: The unit string.

    :returns: True if the unit string represents a combination of SI units,
              False otherwise.
    :rtype: bool
    """
    atomic_unit = "{prefix}?{unit}{power}?".format(
        prefix=PREFIXES, unit=UNITS, power=POWER
    )
    compound_unit = re.compile(
        "({atomic}(\\*|/))+{atomic}".format(atomic=atomic_unit)
    )
    return unit and compound_unit.search(unit)


def scalable(unit_a, unit_b):
    """
    Checks whether units are scalable versions of the same SI unit.
    Method works on two lists and compares the corresponding units in both
    lists.

    :param unit_a: List of unit strings.
    :param unit_b: List of unit strings.

    :returns: True if all corresponding units are scalable.
    :rtype: bool
    """
    if (isinstance(unit_a, Sequence) and isinstance(unit_b, Sequence) and
            not isinstance(unit_a, strings) and
            not isinstance(unit_b, strings)):
        if len(unit_a) != len(unit_b):
            return False
        for a, b in zip(unit_a, unit_b):
            if not scalable(a, b):
                return False
        return True

    if not (is_si(unit_a) and is_si(unit_b)):
        return False

    a_prefix, a_unit, a_power = split(unit_a)
    b_prefix, b_unit, b_power = split(unit_b)
    if a_unit != b_unit or a_power != b_power:
        return False

    return True


def scaling(origin, destination):
    """
    Returns the scaling factor to convert from one unit to another.

    :param origin: The original unit string.
    :param destination: The destination unit string.

    :returns: The scaling factor.
    :rtype: double
    """
    scale = 1.0
    if not scalable(origin, destination):
        raise InvalidUnit(
            "Origin unit and destination unit are not scalable version of the "
            "same SI unit!",
            "nixio.util.scaling"
        )

    org_prefix, org_unit, org_power = split(origin)
    dest_prefix, dest_unit, dest_power = split(destination)

    if org_prefix == dest_prefix and org_power == dest_power:
        return scale
    if not dest_prefix and org_prefix:
        scale = PREFIX_FACTORS[org_prefix]
    elif not org_prefix and dest_prefix:
        scale = 1.0 / PREFIX_FACTORS[dest_prefix]
    elif not org_prefix and not dest_prefix:
        scale = PREFIX_FACTORS[org_prefix] / PREFIX_FACTORS[dest_prefix]

    if org_power:
        power = int(org_power)
        scale **= power

    return scale


def split(combined_unit):
    """
    Splits a unit string into magnitude prefix, the base unit, and the power.

    :param combined_unit: The unit string.

    :returns: A tuple of prefix, base unit, and power.
    :rtype: tuple
    """
    prefix_re = "(?P<prefix>{})".format(PREFIXES)
    unit_re = "(?P<unit>{})".format(UNITS)
    power_re = "(?P<power>{})".format(POWER)
    pup = re.compile(prefix_re + unit_re + power_re)
    pu = re.compile(prefix_re + unit_re)
    up = re.compile(unit_re + power_re)
    # u = re.compile(unit_re)
    # p = re.compile(prefix_re)

    match = pup.match(combined_unit)
    if match:
        prefix = match.group("prefix")
        unit = match.group("unit")
        power = match.group("power")[1:]
        return prefix, unit, power

    match = up.match(combined_unit)
    if match:
        prefix = ""
        unit = match.group("unit")
        power = match.group("power")[1:]
        return prefix, unit, power

    match = pu.match(combined_unit)
    if match:
        prefix = match.group("prefix")
        unit = match.group("unit")
        power = ""
        return prefix, unit, power

    prefix = ""
    unit = combined_unit
    power = ""
    return prefix, unit, power


def invert_power(unit):
    prefix, unit, power = split(unit)
    if not power:
        return prefix + unit + "^-1"
    if power[0] == "-":
        power = power[1:]
    else:
        power = "^-" + power
    return prefix + unit + "^" + power


def split_compound(compound_unit):
    """
    Splits a compound unit (like mV/Hz) into the atomic units.

    :param compound_unit: The unit string.

    :returns: A tuple containing the atomic units.
    :rtype: tuple
    """
    opt_pup = re.compile(PREFIXES + "?" + UNITS + POWER + "?")
    match = opt_pup.match(compound_unit)
    sep = ""
    atomic_units = []
    while match and (match.end() < len(match.string)):
        suffix = match.string[match.end():]
        suffix = suffix.replace(" ", "")
        unit = match.group(0)
        if sep == "/":
            atomic_units.append(invert_power(unit))
        else:
            atomic_units.append(unit)
        sep = suffix[0]
        match = opt_pup.match(suffix[1:])
    unit = match.group(0)
    if sep == "/":
        atomic_units.append(invert_power(unit))
    else:
        atomic_units.append(unit)
    return tuple(atomic_units)



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
# -*- coding: utf-8 -*-
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.


class LinkType(object):
    Tagged = "Tagged"
    Untagged = "Untagged"
    Indexed = "Indexed"



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
# -*- coding: utf-8 -*-
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import numpy as np

import h5py
from datetime import datetime
from uuid import uuid4, UUID
from ..exceptions import exceptions
from . import names

from nixio.link_type import LinkType

try:
    unicode = unicode
except NameError:
    unicode = str

strings = (str, bytes)
try:
    strings += (basestring,)
except NameError:
    pass

vlen_str_dtype = h5py.special_dtype(vlen=unicode)


def create_id():
    """
    Creates an ID as used for nix entities.

    :returns: The ID.
    :rtype: str
    """
    return str(uuid4())


def is_uuid(id_):
    try:
        UUID(str(id_))
        return True
    except ValueError:
        return False


def check_entity_name_and_type(name, type_):
    check_entity_name(name)
    check_entity_type(type_)


def check_entity_type(type_):
    if not type_:
        raise ValueError("String provided for entity type is empty!")


def check_entity_name(name):
    if not name:
        raise ValueError("String provided for entity name is empty!")
    if not names.check(name):
        raise ValueError("String provided for entity name is invalid!")


def check_entity_id(id_):
    if not is_uuid(id_):
        raise ValueError("String provided for id is not a valid UUID")


def check_empty_str(string, field_name):
    if not string:
        raise ValueError("String provided is empty! {}".format(field_name))


def check_name_or_id(name_or_id):
    if not name_or_id:
        raise ValueError("String provided for entity name is empty!")


def check_entity_input(entity, raise_exception=True):
    if entity is not None:
        return True
    if raise_exception:
        raise exceptions.UninitializedEntity()
    return False


def now_int():
    """
    Returns the current POSIX time as an integer.

    :return: integer POSIX time
    """
    now = datetime.now() - datetime(1970, 1, 1)
    return int(now.total_seconds())


def time_to_str(t):
    """
    Returns the time represented by the parameter in the format of Boost's
    posix time `to_iso_string` function.

    :param t: integer POSIX time
    :return: string in the form "YYYYMMDDTHHMMSS", where T is the date-time
    separator
    """
    dt = datetime.utcfromtimestamp(t)
    return dt.strftime("%Y%m%dT%H%M%S").encode("utf-8")


def str_to_time(s):
    """
    Returns the POSIX time represented by the given string as an integer.

    :param s: string in the form "YYYYMMDDTHHMMSS", where T is the date-time
    separator
    :return: integer POSIX time
    """
    if isinstance(s, bytes):
        s = s.decode()
    dt = datetime.strptime(s, "%Y%m%dT%H%M%S") - datetime(1970, 1, 1)
    return int(dt.total_seconds())


def check_attr_type(value, type_):
    """
    Checks if a value is of a given type and raises an exception if the check
    fails. The check does not fail if value is None.
    Specifying `str` for type checks against all string types
    (str, bytes, basestring).


    :param value: the value to check
    :param type_: the type to check against
    """
    if type_ is str:
        type_ = strings
    if value is not None and not isinstance(value, type_):
        raise exceptions.InvalidAttrType(type_, value)


def co_to_slice(count, offset):
    """
    Converts an offset-count pair to an h5py compatible slice

    :param count: number of items
    :param offset: offset from start
    :return: slice-like tuple or single index
    """
    sl = []
    for c, o in zip(count, offset):
        sl.append(slice(o, c+o))
    if len(sl) == 1:
        return sl[0]
    else:
        return tuple(sl)


def apply_polynomial(coefficients, origin, data):
    data[:] = data[:] - origin
    if coefficients:
        data[:] = np.polynomial.polynomial.polyval(data, coefficients)


# The following two functions currently behave as capitalised <-> lowercase
# converters, but they are general solutions for alternate implementations of
# LinkType (e.g., enum)
def link_type_to_string(lt):
    if lt == LinkType.Indexed:
        return "indexed"
    elif lt == LinkType.Tagged:
        return "tagged"
    elif lt == LinkType.Untagged:
        return "untagged"
    else:
        raise RuntimeError("Invalid LinkType")


def link_type_from_string(ltstr):
    if ltstr == "indexed":
        return LinkType.Indexed
    elif ltstr == "tagged":
        return LinkType.Tagged
    elif ltstr == "untagged":
        return LinkType.Untagged
    else:
        raise RuntimeError("Invalid string for LinkType")



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from .entity import NamedEntity
from .section import Section


class EntityWithMetadata(NamedEntity):

    def __init__(self, nixparent, h5group):
        super(EntityWithMetadata, self).__init__(nixparent, h5group)
        # TODO: Additional validation for metadata

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_):
        newentity = super(EntityWithMetadata, cls)._create_new(
            nixparent, h5parent, name, type_
        )
        return newentity

    @property
    def metadata(self):
        """
        Associated metadata of the entity. Sections attached to the entity
        via this attribute can provide additional annotations. This is an
        optional read-write property, and can be None if no metadata is
        available.

        :type: Section
        """
        if "metadata" in self._h5group:
            return Section(None, self._h5group.open_group("metadata"))
        else:
            return None

    @metadata.setter
    def metadata(self, sect):
        if not isinstance(sect, Section):
            raise TypeError("Error setting metadata to {}. Not a Section."
                            .format(sect))
        self._h5group.create_link(sect, "metadata")



import os, sys, shutil

from textwrap import dedent
import tempfile

import distutils.sysconfig
import distutils.ccompiler
from distutils.errors import CompileError, LinkError


def check_nix(library_dirs=(), include_dirs=(), compile_args=()):
    """
    Check if NIX is available by trying to compile a tiny program.
    """

    cpp_code = dedent("""
    #include <nix.hpp>

    int main() {
        return 0;
    }
    """)

    tmpdir = tempfile.mkdtemp()
    bin_file_name = os.path.join(tmpdir, "check_nix")
    file_name = bin_file_name + '.cpp'
    with open(file_name, 'w') as cfile:
        cfile.write(cpp_code)

    compiler = distutils.ccompiler.new_compiler()
    assert isinstance(compiler, distutils.ccompiler.CCompiler)

    distutils.sysconfig.customize_compiler(compiler)
    compiler.set_library_dirs(library_dirs)
    compiler.set_include_dirs(include_dirs)

    stderr = os.dup(sys.stderr.fileno())
    errfile = open(os.path.join(tmpdir, "check_nix.err"), 'w')
    # redirect stderr to file
    os.dup2(errfile.fileno(), sys.stderr.fileno())
    try:
        compiler.compile([file_name], output_dir=tmpdir,
                         extra_postargs=compile_args)
    except (CompileError, LinkError):
        ret_val = False
    else:
        ret_val = True
    # restore stderr
    os.dup2(stderr, sys.stderr.fileno())
    errfile.close()
    shutil.rmtree(tmpdir)
    return ret_val


if __name__ == '__main__':
    nix_inc_dir = os.getenv('NIX_INCDIR', '/usr/local/include')
    nix_lib_dir = os.getenv('NIX_LIBDIR', '/usr/local/lib')
    res = check_nix([nix_lib_dir], [nix_inc_dir])
    if res:
        print("Found NIX.")
    else:
        print("NIX not found.")





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Copyright © 2014 German Neuroinformatics Node (G-Node)

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted under the terms of the BSD License. See
 LICENSE file in the root of the Project.

 Author: Jan Grewe <jan.grewe@g-node.org>

 This tutorial shows how to store image data in nix-files.
 See https://github.com/G-node/nix/wiki for more information.

 We use the "Lenna" image in this tutorial.

 "Lenna" by Original full portrait: "Playmate of the Month". Playboy
 Magazine. November 1972, photographed by Dwight Hooker.This 512x512
 electronic/mechanical scan of a section of the full portrait:
 Alexander Sawchuk and two others[1] - The USC-SIPI image
 database. Via Wikipedia -
 http://en.wikipedia.org/wiki/File:Lenna.png#mediaviewer/File:Lenna.png

"""

import nixio as nix
import numpy as np
import Image as img
import matplotlib.pyplot as plt


def load_image():
    image = img.open('lenna.png')
    pix = np.array(image)
    channels = list(image.mode)
    return pix, channels


def draw_rect(img_data, position, extent):
    img_data[position[0]:position[0] + extent[0], position[1], :] = 255
    img_data[position[0]:position[0] + extent[0], position[1] + extent[1], :] = 255
    img_data[position[0], position[1]:position[1] + extent[1], :] = 255
    img_data[position[0] + extent[0], position[1]:position[1] + extent[1], :] = 255
    return img_data


def plot_data(tag):
    data_array = tag.references[0]
    img_data = data_array[:]
    img_data = np.array(img_data, dtype='uint8')
    positions_data = tag.positions[:]
    extents_data = tag.extents[:]
    for i in range(positions.data_extent[0]):
        img_data = draw_rect(
            img_data, positions_data[i, :], extents_data[i, :])
    new_img = img.fromarray(img_data)
    new_img.show()


def plot_roi_data(tag):
    position_count = tag.positions.shape[0]
    for p in range(position_count):
        roi_data = tag.retrieve_data(p, 0)[:]
        roi_data = np.array(roi_data, dtype='uint8')
        ax = plt.gcf().add_subplot(position_count, 1, p)
        image = img.fromarray(roi_data)
        ax.imshow(image)
    plt.savefig('retrieved_rois.png')
    plt.show()


if __name__ == '__main__':
    img_data, channels = load_image()
    # create a new file overwriting any existing content
    file_name = 'multiple_roi.h5'
    file = nix.File.open(file_name, nix.FileMode.Overwrite)

    # create a 'Block' that represents a grouping object. Here, the recording session.
    # it gets a name and a type
    block = file.create_block("block name", "nix.session")

    # create a 'DataArray' to take the sinewave, add some information about
    # the signal
    data = block.create_data_array("lenna", "nix.image.rgb", data=img_data)
    # add descriptors for width, height and channels
    height_dim = data.append_sampled_dimension(1)
    height_dim.label = "height"
    width_dim = data.append_sampled_dimension(1)
    width_dim.label = "width"
    color_dim = data.append_set_dimension()
    color_dim.labels = channels

    # some space for three regions-of-interest
    roi_starts = np.zeros((3, 3))
    roi_starts[0, :] = [250, 245, 0]
    roi_starts[1, :] = [250, 315, 0]
    roi_starts[2, :] = [340, 260, 0]

    roi_extents = np.zeros((3, 3))
    roi_extents[0, :] = [30, 45, 3]
    roi_extents[1, :] = [30, 40, 3]
    roi_extents[2, :] = [25, 65, 3]

    # create the positions DataArray
    positions = block.create_data_array("ROI positions", "nix.positions", data=roi_starts)
    positions.append_set_dimension()  # these can be empty
    positions.append_set_dimension()

    # create the extents DataArray
    extents = block.create_data_array("ROI extents", "nix.extents", data=roi_extents)
    extents.append_set_dimension()
    extents.append_set_dimension()

    # create a MultiTag
    multi_tag = block.create_multi_tag("Regions of interest", "nix.roi", positions)
    multi_tag.extents = extents
    multi_tag.references.append(data)

    # let's plot the data from the stored information
    plot_data(multi_tag)
    plot_roi_data(multi_tag)
    file.close()



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)

from ..group import GroupMixin
from .entity_with_sources import EntityWithSources
from .data_array import DataArray
from .tag import Tag
from .multi_tag import MultiTag

from . import util


class Group(EntityWithSources, GroupMixin):

    def __init__(self, nixparent, h5group):
        super(Group, self).__init__(nixparent, h5group)

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_):
        newentity = super(Group, cls)._create_new(nixparent, h5parent,
                                                  name, type_)
        return newentity

    # DataArray
    def _get_data_array_by_id(self, id_or_name):
        data_arrays = self._h5group.open_group("data_arrays")
        if not util.is_uuid(id_or_name):
            id_or_name = self._parent.data_arrays[id_or_name].id
        # Using get_by_name - linked entries use id as name in backend
        return DataArray(self._parent, data_arrays.get_by_name(id_or_name))

    def _get_data_array_by_pos(self, pos):
        data_arrays = self._h5group.open_group("data_arrays")
        return DataArray(self._parent, data_arrays.get_by_pos(pos))

    def _delete_data_array_by_id(self, id_):
        data_arrays = self._h5group.open_group("data_arrays")
        data_arrays.delete(id_)

    def _data_array_count(self):
        return len(self._h5group.open_group("data_arrays"))

    def _add_data_array_by_id(self, id_or_name):
        if id_or_name not in self._parent.data_arrays:
            raise RuntimeError("Group._add_data_array_by_id: "
                               "DataArray not found in Block!")
        target = self._parent.data_arrays[id_or_name]
        data_arrays = self._h5group.open_group("data_arrays")
        data_arrays.create_link(target, target.id)

    def _has_data_array_by_id(self, id_or_name):
        data_arrays = self._h5group.open_group("data_arrays")
        return data_arrays.has_by_id(id_or_name)

    # MultiTag
    def _get_multi_tag_by_id(self, id_or_name):
        multi_tags = self._h5group.open_group("multi_tags")
        if not util.is_uuid(id_or_name):
            id_or_name = self._parent.multi_tags[id_or_name].id
        # Using get_by_name - linked entries use id as name in backend
        return MultiTag(self._parent, multi_tags.get_by_name(id_or_name))

    def _get_multi_tag_by_pos(self, pos):
        multi_tags = self._h5group.open_group("multi_tags")
        return MultiTag(self._parent, multi_tags.get_by_pos(pos))

    def _delete_multi_tag_by_id(self, id_):
        multi_tags = self._h5group.open_group("multi_tags")
        multi_tags.delete(id_)

    def _multi_tag_count(self):
        return len(self._h5group.open_group("multi_tags"))

    def _add_multi_tag_by_id(self, id_or_name):
        if id_or_name not in self._parent.multi_tags:
            raise RuntimeError("Group._add_multi_tag_by_id: "
                               "MultiTag not found in Block!")
        target = self._parent.multi_tags[id_or_name]
        multi_tags = self._h5group.open_group("multi_tags")
        multi_tags.create_link(target, target.id)

    def _has_multi_tag_by_id(self, id_or_name):
        multi_tags = self._h5group.open_group("multi_tags")
        return multi_tags.has_by_id(id_or_name)

    # Tag
    def _get_tag_by_id(self, id_or_name):
        tags = self._h5group.open_group("tags")
        if not util.is_uuid(id_or_name):
            id_or_name = self._parent.tags[id_or_name].id
        # Using get_by_name - linked entries use id as name in backend
        return Tag(self._parent, tags.get_by_name(id_or_name))

    def _get_tag_by_pos(self, pos):
        tags = self._h5group.open_group("tags")
        return Tag(self._parent, tags.get_by_pos(pos))

    def _delete_tag_by_id(self, id_):
        tags = self._h5group.open_group("tags")
        tags.delete(id_)

    def _tag_count(self):
        return len(self._h5group.open_group("tags"))

    def _add_tag_by_id(self, id_or_name):
        if id_or_name not in self._parent.tags:
            raise RuntimeError("Group._add_tag_by_id: "
                               "Tag not found in Block!")
        target = self._parent.tags[id_or_name]
        tags = self._h5group.open_group("tags")
        tags.create_link(target, target.id)

    def _has_tag_by_id(self, id_or_name):
        tags = self._h5group.open_group("tags")
        return tags.has_by_id(id_or_name)



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from nixio.tag import ReferenceProxyList, FeatureProxyList


class MultiTagMixin(object):

    @property
    def references(self):
        """
        A property containing all data arrays referenced by the tag. Referenced
        data arrays can be obtained by index or their id. References can be
        removed from the list, removing a referenced DataArray will not remove
        it from the file. New references can be added using the append method
        of the list.
        This is a read only attribute.

        :type: RefProxyList of DataArray
        """
        if not hasattr(self, "_references"):
            setattr(self, "_references", ReferenceProxyList(self))
        return self._references

    @property
    def features(self):
        """
        A property containing all features of the tag. Features can be obtained
        via their index or their id. Features can be deleted from the list.
        Adding new features to the tag is done using the create_feature method.
        This is a read only attribute.

        :type: ProxyList of Feature.
        """
        if not hasattr(self, "_features"):
            setattr(self, "_features", FeatureProxyList(self))
        return self._features



# Copyright (c) 2015, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)
from numbers import Number, Integral, Real

import numpy as np

strings = (str, bytes)
try:
    strings += (basestring,)
except NameError:
    pass

bools = (bool, np.bool_)

valid_types = (Number, strings, bools)


class DataType(object):
    UInt8 = np.uint8
    UInt16 = np.uint16
    UInt32 = np.uint32
    UInt64 = np.uint64
    Int8 = np.int8
    Int16 = np.int16
    Int32 = np.int32
    Int64 = np.int64
    Float = np.float_
    Double = np.double
    String = np.string_
    Bool = np.bool_

    @classmethod
    def get_dtype(cls, value):
        if isinstance(value, bools):
            return cls.Bool
        elif isinstance(value, Integral):
            return cls.Int64
        elif isinstance(value, Real):
            return cls.Float
        elif isinstance(value, strings):
            return cls.String
        else:
            raise ValueError("Unknown type for value {}".format(value))

    @classmethod
    def is_numeric_dtype(cls, dtype):
        return (dtype == cls.Int8 or
                dtype == cls.Int16 or
                dtype == cls.Int32 or
                dtype == cls.Int64 or
                dtype == cls.UInt8 or
                dtype == cls.UInt16 or
                dtype == cls.UInt32 or
                dtype == cls.UInt64 or
                dtype == cls.Float or
                dtype == cls.Double)


class Value(object):
    def __init__(self, value):
        if isinstance(value, valid_types):
            self.value = value
        else:
            raise TypeError("Invalid value type: {}".format(type(value)))
        self.uncertainty = 0
        self.reference = ""
        self.filename = ""
        self.encoder = ""
        self.checksum = ""
        self.data_type = DataType.get_dtype(value)

    def __str__(self):
        return "Value{{[{dtype}] {value}}}".format(dtype=self.data_type,
                                                   value=self.value)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if hasattr(other, "value"):
            return self.value == other.value
        else:
            return self.value == other

    def to_string(self, unit=""):
        value_str = ""
        err = "" if self.uncertainty == 0.0 else "+- %.3f" % self.uncertainty
        value_str = "%s%s%s" % (str(self.value), err, unit)
        return value_str



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import nixio.util.find as finders
from nixio.util.proxy_list import ProxyList

try:
    from sys import maxint
except:
    from sys import maxsize as maxint


class BlockProxyList(ProxyList):

    def __init__(self, obj):
        super(BlockProxyList, self).__init__(obj, "_block_count",
                                             "_get_block_by_id",
                                             "_get_block_by_pos",
                                             "_delete_block_by_id")


class SectionProxyList(ProxyList):

    def __init__(self, obj):
        super(SectionProxyList, self).__init__(obj, "_section_count",
                                               "_get_section_by_id",
                                               "_get_section_by_pos",
                                               "_delete_section_by_id")


class FileMixin(object):

    @property
    def blocks(self):
        """
        A property containing all blocks of a file. Blocks can be obtained by
        their id or their index. Blocks can be deleted from the list, when a
        block is deleted all its content (data arrays, tags and sources) will
        be also deleted from the file. Adding new Block is done via the
        create_block method of File. This is a read-only attribute.

        :type: ProxyList of Block entities.
        """
        if not hasattr(self, "_blocks"):
            setattr(self, "_blocks", BlockProxyList(self))
        return self._blocks

    def find_sections(self, filtr=lambda _: True, limit=None):
        """
        Get all sections and their child sections recursively.

        This method traverses the trees of all sections. The traversal is
        accomplished via breadth first and can be limited in depth. On each
        node or section a filter is applied. If the filter returns true the
        respective section will be added to the result list.
        By default a filter is used that accepts all sections.

        :param filtr: A filter function
        :type filtr:  function
        :param limit: The maximum depth of traversal
        :type limit:  int

        :returns: A list containing the matching sections.
        :rtype: list of Section
        """
        if limit is None:
            limit = maxint
        return finders._find_sections(self, filtr, limit)

    @property
    def sections(self):
        """
        A property containing all root sections of a file. Specific root
        sections can be obtained by their id or their index. Sections can be
        deleted from this list. Notice: when a section is deleted all its child
        section and properties will be removed too. Adding a new Section is
        done via the crate_section method of File.
        This is a read-only property.

        :type: ProxyList of Section entities.
        """
        if not hasattr(self, "_sections"):
            setattr(self, "_sections", SectionProxyList(self))
        return self._sections



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.


class DimensionType(object):
    Sample = "sample"
    Range = "range"
    Set = "set"



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from nixio.util.proxy_list import RefProxyList


class RefSourceProxyList(RefProxyList):

    def __init__(self, obj):
        super(RefSourceProxyList, self).__init__(
            obj, "_source_count", "_get_source_by_id", "_get_source_by_pos",
            "_remove_source_by_id", "_add_source_by_id"
        )


_sources_doc = """
Getter for sources.
"""


def _get_sources(self):
    if not hasattr(self, "_sources"):
        setattr(self, "_sources", RefSourceProxyList(self))
    return self._sources


class EntityWithSourcesMixin(object):

    sources = property(_get_sources, None, None, _sources_doc)



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
import os
import json

here = os.path.dirname(__file__)

with open(os.path.join(here, "info.json")) as infofile:
    infodict = json.load(infofile)


VERSION = infodict["VERSION"]
STATUS = infodict["STATUS"]
RELEASE = infodict["RELEASE"]
AUTHOR = infodict["AUTHOR"]
COPYRIGHT = infodict["COPYRIGHT"]
CONTACT = infodict["CONTACT"]
BRIEF = infodict["BRIEF"]
HOMEPAGE = infodict["HOMEPAGE"]



import os
import sys
import re
from subprocess import check_output, call, CalledProcessError, DEVNULL
from difflib import Differ
import requests
import json

red_begin = "\033[91m"
red_end = "\033[0m"
bold_begin = "\033[1m"
bold_end = red_end

gitroot = check_output(["git", "rev-parse",
                        "--show-toplevel"]).decode().strip()


def die(msg):
    print(msg, file=sys.stderr)
    sys.exit(1)


def check_git_status():
    ret = call(["git", "diff-index", "--quiet", "HEAD", "--"])
    if ret > 0:
        print("{}There are unstaged changes in the working tree.{}\n"
              "Are you sure you want to continue?".format(
                  red_begin, red_end
              ))
        wait_for_ret()


def chdir():
    scriptpath = sys.argv[0]
    scriptdir = os.path.dirname(scriptpath) or "."
    os.chdir(scriptdir)


def parse_version_string(ver):
    try:
        ver = tuple(map(int, ver.split(".")))
        if len(ver) != 3:
            raise ValueError
        return ver
    except ValueError:
        die("Invalid version string. Please provide a version of the form "
            "X.Y.Z (all integers).")


def wait_for_ret():
    try:
        input("Hit return to continue or ^C to cancel ...")
    except KeyboardInterrupt:
        die("\nCancelled")


def diff_lines(old, new):
    d = Differ()
    diff = list(d.compare(old, new))
    for line in diff[:]:
        if line.startswith(" "):
            diff.remove(line)

    return diff


def update_readme():
    branch = check_output(["git", "rev-parse",
                           "--abbrev-ref", "HEAD"]).decode().strip()
    print("Current branch is '{}'".format(branch))
    wait_for_ret()

    readmefn = os.path.join(gitroot, "README.rst")

    with open(readmefn) as oldreadme:
        oldrmtext = oldreadme.readlines()

    newrmtext = []
    for line in oldrmtext:
        if "travis-ci.org" in line or "coveralls.io" in line:
            line = line.replace("master", branch)
        newrmtext.append(line)

    diff = diff_lines(oldrmtext, newrmtext)

    if not diff:
        print("No changes required in README.")
        wait_for_ret()
        return False

    print("".join(diff))
    print("{}The above changes will be written to README.rst{}".format(
        red_begin, red_end
    ))
    wait_for_ret()
    with open(readmefn, "w") as readme:
        readme.writelines(newrmtext)
    return True


def update_info(newver):
    infofn = os.path.join(gitroot, "nixio", "info.json")
    with open(infofn) as infofile:
        infodict = json.load(infofile)

    verstring = infodict["VERSION"]
    newverstring = re.sub("[0-9\.a-z]+", newver, verstring)

    if newverstring == verstring:
        print("No changes required in info.json")
        wait_for_ret()
        return False

    oldrelstring = infodict["RELEASE"]
    newrelstring = "{} Release".format(newverstring)

    print("VERSION: '{}' → '{}'".format(verstring, newverstring))
    print("RELEASE: '{}' → '{}'".format(oldrelstring, newrelstring))
    print("{}The above changes will be written to info.json{}".format(
        red_begin, red_end
    ))
    wait_for_ret()
    with open(infofn, "w") as infofile:
        infodict["VERSION"] = newverstring
        infodict["RELEASE"] = newrelstring
        json.dump(infodict, infofile)

    return True


def update_ci_confs(newver):
    change = False
    nixbranch = ".".join(newver.split(".")[:-1])
    travisfn = os.path.join(gitroot, ".travis.yml")
    with open(travisfn) as travisconf:
        oldconf = travisconf.readlines()

    newconf = []
    for line in oldconf:
        if "NIX_BRANCH" in line:
            line = re.sub("1\.[0-9\.]+[0-9\.a-z]+(dev){0,1}", nixbranch, line)
            line = line.replace("master", nixbranch)
        newconf.append(line)

    diff = diff_lines(oldconf, newconf)

    if len(diff) == 0:
        print("No changes required in .travis.yml")
        wait_for_ret()
    else:
        print("".join(diff))
        print("{}The above changes will be written to .travis.yml{}".format(
            red_begin, red_end
        ))
        wait_for_ret()
        with open(travisfn, "w") as travisconf:
            travisconf.writelines(newconf)
        change = True

    # Appveyor is currently not building NIX so let's not bother changing

    # appveyorfn = os.path.join(gitroot, "appveyor.yml")
    # with open(appveyorfn) as avconf:
    #     oldconf = avconf.readlines()

    # newconf = []
    # for line in oldconf:
    #     if "NIX_VERSION" in line:
    #         line = re.sub("'[1-9\.a-z]+'", "'" + newver + "'", line)
    #     newconf.append(line)

    # diff = diff_lines(oldconf, newconf)

    # if len(diff) == 0:
    #     print("No changes required in appveyor.yml")
    #     wait_for_ret()
    # else:
    #     print("".join(diff))
    #     print("{}The above changes will be written to appveyor.yml{}".format(
    #         red_begin, red_end
    #     ))
    #     wait_for_ret()
    #     with open(appveyorfn, "w") as avconf:
    #         avconf.writelines(newconf)
    #     change = True

    return change


def tag_head(newverstr):
    try:
        tagrev = check_output(["git", "rev-parse", newverstr], stderr=DEVNULL)
        headrev = check_output(["git", "rev-parse", "HEAD"])
        if tagrev != headrev:
            die("Tag or object named {} already exists "
                "and is not current HEAD.".format(newverstr))
        tagexists = True
    except CalledProcessError:
        tagexists = False

    if tagexists:
        print("HEAD already tagged with {}".format(newverstr))
        return

    print("{}Tagging current commit with '{}'{}".format(
        red_begin, newverstr, red_end))
    print("[{}$ git tag {}{}]".format(red_begin, newverstr, red_end))
    wait_for_ret()
    ret = call(["git", "tag", newverstr])
    if ret > 0:
        die("Error during 'git tag' operation.")


def print_appveyor_urls():
    print("Hit return to print the URLs for the latest wheels on Appveyor.")
    print("If the build is still running, you can wait until it's finished.")
    wait_for_ret()

    apiurl = "https://ci.appveyor.com/api/"
    account = "achilleas-k"
    project_name = "nixpy-um2sy"

    url = os.path.join(apiurl, "projects", account, project_name)
    r = requests.get(url)

    projects = json.loads(r.text)
    build = projects["build"]
    for job in build["jobs"]:
        if job["status"] == "success":
            artifacts_url = os.path.join(apiurl, "buildjobs", job["jobId"],
                                         "artifacts")
            r = requests.get(artifacts_url)
            artifacts = json.loads(r.text)
            print(job["name"])
            for a in artifacts:
                print(os.path.join(apiurl, "buildjobs", job["jobId"],
                                   "artifacts", a["fileName"]))


def main():
    check_git_status()

    chdir()

    if len(sys.argv) < 2:
        die("Specify version (X.Y.Z)")

    newverstr = sys.argv[1]
    parse_version_string(newverstr)

    prepline = "Preparing new release: {}".format(newverstr)
    banner = "="*len(prepline) + "\n" + prepline + "\n" + "="*len(prepline)
    banner = bold_begin + banner + bold_end
    print(banner)

    chg = update_readme() | update_info(newverstr) | update_ci_confs(newverstr)
    if chg:
        print("Files have changed. Commit these changes and rerun the script.")
        sys.exit(0)

    check_git_status()  # might have been external changes (?) check again

    tag_head(newverstr)

    print("Creating archive...")
    os.chdir(gitroot)
    ret = call(["python", "setup.py", "sdist"])

    if ret > 0:
        die("Error creating package")

    print("{}Package ready. See 'dist' directory under project root.{}".format(
        bold_begin, bold_end
    ))

    print_appveyor_urls()


if __name__ == "__main__":
    main()
